from typing import Dict, Iterable, Optional

from polyaxon import pkg
from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._flow import V1CompiledOperation, V1KFReplica, V1Plugins, V1PytorchJob
from polyaxon._k8s.converter.base import BaseConverter
from polyaxon._k8s.converter.mixins import PytorchJobMixin
from polyaxon._k8s.custom_resources.kubeflow import get_pytorch_job_custom_resource
from polyaxon._k8s.replica import ReplicaSpec


class PytorchJobConverter(PytorchJobMixin, BaseConverter):
    def get_resource(
        self,
        compiled_operation: V1CompiledOperation,
        artifacts_store: V1Connection,
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        default_sa: Optional[str] = None,
        default_auth: bool = False,
    ) -> Dict:
        job = compiled_operation.run  # type: V1PytorchJob

        def _get_replica(replica: Optional[V1KFReplica]) -> Optional[ReplicaSpec]:
            if not replica:
                return None
            return self.get_replica_resource(
                plugins=plugins,
                environment=replica.environment,
                volumes=replica.volumes or [],
                init=replica.init or [],
                sidecars=replica.sidecars or [],
                container=replica.container,
                artifacts_store=artifacts_store,
                connections=replica.connections or [],
                connection_by_names=connection_by_names,
                secrets=secrets,
                config_maps=config_maps,
                kv_env_vars=kv_env_vars,
                default_sa=default_sa,
                num_replicas=replica.replicas,
            )

        kv_env_vars = compiled_operation.get_env_io()
        plugins = V1Plugins.get_or_create(
            config=compiled_operation.plugins, auth=default_auth
        )
        master = _get_replica(job.master)
        worker = _get_replica(job.worker)
        labels = self.get_labels(version=pkg.VERSION, labels={})

        return get_pytorch_job_custom_resource(
            namespace=self.namespace,
            resource_name=self.resource_name,
            master=master,
            worker=worker,
            termination=compiled_operation.termination,
            clean_pod_policy=job.clean_pod_policy,
            scheduling_policy=job.scheduling_policy,
            elastic_policy=job.elastic_policy,
            n_proc_per_node=job.n_proc_per_node,
            collect_logs=plugins.collect_logs,
            sync_statuses=plugins.sync_statuses,
            notifications=plugins.notifications,
            labels=labels,
            annotations=self.annotations,
        )



import aiofiles

from clipped.utils.json import orjson_dumps
from clipped.utils.paths import check_or_create_path

from polyaxon._contexts import paths as ctx_paths
from polyaxon._flow import V1RunKind
from polyaxon._k8s.logging.async_monitor import get_op_spec
from polyaxon._k8s.manager.async_manager import AsyncK8sManager


async def sync_spec(
    run_uuid: str,
    k8s_manager: AsyncK8sManager,
    run_kind: V1RunKind,
):
    op_spec = await get_op_spec(
        k8s_manager=k8s_manager, run_uuid=run_uuid, run_kind=run_kind
    )
    path_from = ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(run_uuid)
    path_from = "{}/outputs/spec.json".format(path_from)
    check_or_create_path(path_from, is_dir=False)
    async with aiofiles.open(path_from, "w") as filepath:
        await filepath.write(orjson_dumps(op_spec))



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr


class V1EntitiesTags(BaseAllowSchemaModel):
    uuids: Optional[List[UUIDStr]]
    tags: Optional[List[StrictStr]]



from typing import List, Optional

from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_settings_catalog import V1SettingsCatalog


class V1TeamSettings(BaseAllowSchemaModel):
    projects: Optional[List[V1SettingsCatalog]]
    hubs: Optional[List[V1SettingsCatalog]]



import os

from typing import Optional

from clipped.compact.pydantic import ValidationError
from clipped.formatting import Printer
from clipped.utils.bools import to_bool

from polyaxon._env_vars.keys import ENV_KEYS_NO_CONFIG, ENV_KEYS_SET_AGENT
from polyaxon._managers.client import ClientConfigManager
from polyaxon._managers.home import HomeConfigManager
from polyaxon._managers.user import UserConfigManager
from polyaxon._schemas.agent import AgentConfig
from polyaxon._schemas.authentication import AccessTokenConfig
from polyaxon._schemas.cli import CliConfig
from polyaxon._schemas.client import ClientConfig
from polyaxon._schemas.home import HomeConfig
from polyaxon._services.values import PolyaxonServices
from polyaxon.api import LOCALHOST

from polyaxon._managers.ignore import IgnoreConfigManager  # noqa
from polyaxon._managers.run import RunConfigManager  # noqa
from polyaxon._managers.project import ProjectConfigManager  # noqa
from polyaxon._managers.auth import AuthConfigManager  # noqa
from polyaxon._managers.cli import CliConfigManager  # noqa

MIN_TIMEOUT = 1
LONG_REQUEST_TIMEOUT = 3600
HEALTH_CHECK_INTERVAL = 60
SET_AGENT = to_bool(os.environ.get(ENV_KEYS_SET_AGENT, False))

HOME_CONFIG: HomeConfig = HomeConfigManager.get_config_from_env()
AUTH_CONFIG: Optional[AccessTokenConfig] = None
CLIENT_CONFIG: ClientConfig
CLI_CONFIG: Optional[CliConfig] = None
AGENT_CONFIG: Optional[AgentConfig] = None

PolyaxonServices.set_service_name()


def set_home_config(config: Optional[HomeConfig] = None):
    global HOME_CONFIG

    HOME_CONFIG = config or HomeConfigManager.get_config_from_env()


def set_agent_config(config: Optional[AgentConfig] = None):
    from polyaxon._connections import CONNECTION_CONFIG
    from polyaxon._managers.agent import AgentConfigManager

    global AGENT_CONFIG

    # Patch the config with correct home path if available
    AgentConfigManager.set_config_path(HOME_CONFIG.path)

    AGENT_CONFIG = config or AgentConfigManager.get_config_from_env()

    # Set artifacts store name
    AGENT_CONFIG.set_artifacts_store_name()

    # Always sync the connections catalog to the current agent config
    CONNECTION_CONFIG.set_connections_catalog(AGENT_CONFIG.all_connections)


def set_cli_config():
    global CLI_CONFIG

    # Patch the config with correct home path if available
    CliConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        CLI_CONFIG = CliConfigManager.get_config_or_default()
    except (TypeError, ValidationError):
        CliConfigManager.purge()
        Printer.warning("Your CLI configuration was purged!")


def set_client_config():
    global CLIENT_CONFIG

    # Patch the config with correct home path if available
    ClientConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        CLIENT_CONFIG = ClientConfigManager.get_config_from_env()
    except (TypeError, ValidationError):
        ClientConfigManager.purge()
        Printer.warning("Your client configuration was purged!")
        CLIENT_CONFIG = ClientConfigManager.get_config_from_env()


def set_auth_config():
    global AUTH_CONFIG

    # Patch the config with correct home path if available
    AuthConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        AUTH_CONFIG = AuthConfigManager.get_config_from_env()
    except (TypeError, ValidationError):
        AuthConfigManager.purge()
        Printer.warning("Your auth configuration was purged!")

    # Patch the config with correct home path if available
    UserConfigManager.set_config_path(HOME_CONFIG.path)

    try:
        UserConfigManager.get_config_or_default()
    except (TypeError, ValidationError):
        UserConfigManager.purge()
        Printer.warning("Your user configuration was purged!")


if not to_bool(os.environ.get(ENV_KEYS_NO_CONFIG, False)):
    set_auth_config()
    set_client_config()
    if PolyaxonServices.is_agent() or SET_AGENT:
        set_agent_config()
else:
    # Patch the config with correct home path if available
    ClientConfigManager.set_config_path(HOME_CONFIG.path)

    CLIENT_CONFIG = ClientConfigManager.CONFIG(host=LOCALHOST)



from typing import Dict, List, Optional

from clipped.compact.pydantic import Extra, Field, StrictInt, StrictStr

from polyaxon._deploy.schemas.celery import CeleryConfig
from polyaxon._deploy.schemas.service_types import ServiceTypes
from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._schemas.services import BaseServiceConfig


class DeploymentService(BaseServiceConfig):
    enabled: Optional[bool]
    replicas: Optional[StrictInt]
    concurrency: Optional[StrictInt]
    per_core: Optional[bool] = Field(alias="perCore")
    scheme: Optional[StrictStr]


class WorkerServiceConfig(DeploymentService):
    celery: Optional[CeleryConfig]


class AgentServiceConfig(DeploymentService):
    instance: Optional[StrictStr]
    token: Optional[StrictStr]
    watch_cluster: Optional[bool] = Field(alias="watchCluster")
    additional_namespaces: Optional[List[str]] = Field(alias="additionalNamespaces")
    is_replica: Optional[bool] = Field(alias="isReplica")


class OperatorServiceConfig(DeploymentService):
    skip_crd: Optional[bool] = Field(alias="skipCRD")


class BaseService(BaseSchemaModel):
    name: Optional[StrictStr]
    type: Optional[ServiceTypes]
    port: Optional[StrictInt]
    target_port: Optional[StrictInt] = Field(alias="targetPort")
    node_port: Optional[StrictInt] = Field(alias="nodePort")
    annotations: Optional[Dict]

    class Config:
        extra = Extra.allow


class ApiServiceConfig(DeploymentService):
    service: Optional[BaseService]


class HooksConfig(DeploymentService):
    load_fixtures: Optional[bool] = Field(alias="loadFixtures")
    tables: Optional[bool] = Field(alias="tables")
    sync_db: Optional[bool] = Field(alias="syncdb")
    admin_user: Optional[bool] = Field(alias="adminUser")


class ThirdPartyService(DeploymentService):
    persistence: Optional[Dict]

    class Config:
        extra = Extra.allow


class PostgresqlConfig(ThirdPartyService):
    auth: Optional[Dict]
    conn_max_age: Optional[StrictInt] = Field(alias="connMaxAge")


class RedisConfig(ThirdPartyService):
    image: Optional[Dict]  # type: ignore[assignment]
    non_broker: Optional[bool] = Field(alias="nonBroker")
    use_password: Optional[bool] = Field(alias="usePassword")
    auth: Optional[Dict]


class RabbitmqConfig(ThirdPartyService):
    auth: Optional[Dict]


class ExternalService(BaseSchemaModel):
    user: Optional[StrictStr]
    password: Optional[StrictStr]
    host: Optional[StrictStr]
    port: Optional[StrictInt]
    database: Optional[StrictStr]
    use_password: Optional[bool] = Field(alias="usePassword")
    conn_max_age: Optional[StrictInt] = Field(alias="connMaxAge")
    pgbouncer: Optional[Dict]
    options: Optional[Dict]
    use_resolver: Optional[bool] = Field(alias="useResolver")
    corporate_proxy: Optional[StrictStr] = Field(alias="corporateProxy")


class ExternalBackend(BaseSchemaModel):
    enabled: Optional[bool]
    backend: Optional[StrictStr]
    options: Optional[Dict]


class AuthServicesConfig(BaseSchemaModel):
    github: Optional[ExternalBackend]
    gitlab: Optional[ExternalBackend]
    bitbucket: Optional[ExternalBackend]
    google: Optional[ExternalBackend]
    okta: Optional[ExternalBackend]
    onelogin: Optional[ExternalBackend]
    azuread: Optional[ExternalBackend]


class ExternalServicesConfig(BaseSchemaModel):
    redis: Optional[ExternalService]
    rabbitmq: Optional[ExternalService]
    postgresql: Optional[ExternalService]
    gateway: Optional[ExternalService]
    api: Optional[ExternalService]
    transactions: Optional[ExternalBackend]
    analytics: Optional[ExternalBackend]
    metrics: Optional[ExternalBackend]
    errors: Optional[ExternalBackend]
    auth: Optional[AuthServicesConfig]
    allowed_versions: Optional[List[StrictStr]] = Field(alias="allowedVersions")



from typing import Dict, List, Optional, Tuple

from kubernetes_asyncio import client, config
from kubernetes_asyncio.client import Configuration
from kubernetes_asyncio.client.rest import ApiException

from polyaxon._k8s.manager.base import BaseK8sManager
from polyaxon._k8s.monitor import is_pod_running
from polyaxon.exceptions import PolyaxonK8sError
from polyaxon.logger import logger


class AsyncK8sManager(BaseK8sManager):
    CLIENT = client

    @staticmethod
    async def load_config(
        in_cluster: bool = False, k8s_config: Optional[bool] = None
    ) -> Configuration:
        if not k8s_config:
            if in_cluster:
                config.load_incluster_config()
            else:
                await config.load_kube_config()
        return Configuration.get_default_copy()

    async def setup(self, k8s_config: Optional[Configuration] = None):
        if not k8s_config:
            if self.in_cluster:
                config.load_incluster_config()
            else:
                await config.load_kube_config()
            self.api_client = client.api_client.ApiClient()
        else:
            self.api_client = client.api_client.ApiClient(configuration=k8s_config)

    async def close(self):
        if self.api_client:
            await self.api_client.close()

    async def get_version(self, reraise: bool = False):
        try:
            version = await self.k8s_version_api.get_code()
            return version.to_dict()
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e

    async def get_pod(
        self, name, reraise: bool = False, namespace: str = None
    ) -> Optional[client.V1Pod]:
        try:
            return await self.k8s_api.read_namespaced_pod(  # type: ignore[attr-defined]
                name=name, namespace=namespace or self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    async def is_pod_running(
        self, pod_id: str, container_id: str, namespace: str = None
    ) -> bool:
        event = await self.k8s_api.read_namespaced_pod_status(pod_id, namespace=namespace or self.namespace)  # type: ignore[attr-defined]
        return is_pod_running(event, container_id)

    async def _list_namespace_resource(
        self, resource_api, reraise: bool = False, namespace: str = None, **kwargs
    ) -> List:
        try:
            res = await resource_api(namespace=namespace or self.namespace, **kwargs)
            return [p for p in res.items]
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return []

    async def list_pods(
        self, reraise: bool = False, namespace: str = None, **kwargs
    ) -> List[client.V1Pod]:
        return await self._list_namespace_resource(
            resource_api=self.k8s_api.list_namespaced_pod,  # type: ignore[attr-defined]
            reraise=reraise,
            namespace=namespace,
            **kwargs,
        )

    async def list_jobs(
        self, reraise: bool = False, namespace: str = None, **kwargs
    ) -> List[client.V1Job]:
        return await self._list_namespace_resource(
            resource_api=self.k8s_batch_api.list_namespaced_job,  # type: ignore[attr-defined]
            reraise=reraise,
            namespace=namespace,
            **kwargs,
        )

    async def list_custom_objects(
        self,
        group,
        version,
        plural,
        reraise: bool = False,
        namespace: str = None,
        **kwargs
    ) -> List:
        return await self._list_namespace_resource(
            resource_api=self.k8s_custom_object_api.list_namespaced_custom_object,  # type: ignore[attr-defined]
            reraise=reraise,
            group=group,
            version=version,
            plural=plural,
            namespace=namespace,
            **kwargs,
        )

    async def list_services(
        self, reraise: bool = False, namespace: str = None, **kwargs
    ) -> List[client.V1Service]:
        return await self._list_namespace_resource(
            resource_api=self.k8s_api.list_namespaced_service,  # type: ignore[attr-defined]
            reraise=reraise,
            namespace=namespace,
            **kwargs,
        )

    async def list_deployments(
        self, reraise: bool = False, namespace: str = None, **kwargs
    ) -> List[client.V1Deployment]:
        return await self._list_namespace_resource(
            resource_api=self.k8s_apps_api.list_namespaced_deployment,  # type: ignore[attr-defined]
            reraise=reraise,
            namespace=namespace,
            **kwargs,
        )

    async def create_custom_object(
        self,
        name: str,
        group: str,
        version: str,
        plural: str,
        body: Dict,
        namespace: str = None,
    ) -> Dict:
        resp = await self.k8s_custom_object_api.create_namespaced_custom_object(  # type: ignore[attr-defined]
            group=group,
            version=version,
            plural=plural,
            namespace=namespace or self.namespace,
            body=body,
        )
        logger.debug("Custom object `{}` was created".format(name))
        return resp

    async def update_custom_object(
        self,
        name: str,
        group: str,
        version: str,
        plural: str,
        body: Dict,
        namespace: str = None,
    ) -> Dict:
        resp = await self.k8s_custom_object_api.patch_namespaced_custom_object(  # type: ignore[attr-defined]
            name=name,
            group=group,
            version=version,
            plural=plural,
            namespace=namespace or self.namespace,
            body=body,
            _content_type="application/merge-patch+json",
        )
        logger.debug("Custom object `{}` was patched".format(name))
        return resp

    async def create_or_update_custom_object(
        self,
        name: str,
        group: str,
        version: str,
        plural: str,
        body: Dict,
        reraise: bool = False,
        namespace: str = None,
    ) -> Tuple[Dict, bool]:
        try:
            create = await self.create_custom_object(
                name=name,
                group=group,
                version=version,
                plural=plural,
                body=body,
                namespace=namespace,
            )
            return create, True

        except ApiException as e_create:
            try:
                update = await self.update_custom_object(
                    name=name,
                    group=group,
                    version=version,
                    plural=plural,
                    body=body,
                    namespace=namespace,
                )
                return update, False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError(
                        "Connection error: creation %s - update %s" % (e_create, e)
                    ) from e
                else:
                    logger.error("K8S error: {}".format(e))
        return {}, False

    async def get_custom_object(
        self,
        name: str,
        group: str,
        version: str,
        plural: str,
        reraise: bool = False,
        namespace: str = None,
    ) -> Optional[Dict]:
        try:
            return await self.k8s_custom_object_api.get_namespaced_custom_object(
                name=name,
                group=group,
                version=version,
                plural=plural,
                namespace=namespace or self.namespace,
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    async def delete_custom_object(
        self,
        name: str,
        group: str,
        version: str,
        plural: str,
        reraise: bool = False,
        namespace: str = None,
    ):
        try:
            await self.k8s_custom_object_api.delete_namespaced_custom_object(
                name=name,
                group=group,
                version=version,
                plural=plural,
                namespace=namespace or self.namespace,
                body=client.V1DeleteOptions(),
            )
            logger.debug("Custom object `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Custom object `{}` was not found".format(name))



import os

from typing import Optional

from clipped.config.contexts import get_project_path, get_temp_path
from clipped.utils.enums import get_enum_value

from polyaxon._env_vars.keys import (
    ENV_KEYS_ARCHIVES_ROOT,
    ENV_KEYS_ARTIFACTS_ROOT,
    ENV_KEYS_CONTEXT_ROOT,
    ENV_KEYS_OFFLINE_ROOT,
)

CONTEXT_RELATED_RUNS = "_related_runs"

# Local contexts
CONTEXT_LOCAL_LINEAGES = "lineages.plx.json"
CONTEXT_LOCAL_CONTENT = "content.plx.json"
CONTEXT_LOCAL_README = "readme.plx.md"
CONTEXT_LOCAL_POLYAXONFILE = "polyaxonfile.plx.md"
CONTEXT_LOCAL_PROJECT = "project.plx.json"
CONTEXT_LOCAL_RUN = "run.plx.json"
CONTEXT_LOCAL_VERSION = "version.plx.json"

CONTEXT_ROOT = os.environ.get(ENV_KEYS_CONTEXT_ROOT, "/plx-context")
CONTEXT_MOUNT_CONFIGS = "{}/.configs".format(CONTEXT_ROOT)
CONTEXT_MOUNT_AUTH = "{}/.auth".format(CONTEXT_MOUNT_CONFIGS)
CONTEXT_MOUNT_ARTIFACTS = "{}/artifacts".format(CONTEXT_ROOT)
CONTEXT_MOUNT_FILE_WATCHER = "{}/.fs".format(CONTEXT_MOUNT_ARTIFACTS)
CONTEXT_MOUNT_ARTIFACTS_FORMAT = "{}/{{}}".format(CONTEXT_MOUNT_ARTIFACTS)
CONTEXT_MOUNT_ARTIFACTS_RELATED = CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(
    CONTEXT_RELATED_RUNS
)
CONTEXT_MOUNT_ARTIFACTS_RELATED_FORMAT = "{}/{{}}".format(
    CONTEXT_MOUNT_ARTIFACTS_RELATED
)
CONTEXT_MOUNT_RUN_ASSETS_FORMAT = "{}/assets".format(CONTEXT_MOUNT_ARTIFACTS_FORMAT)
CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT = "{}/outputs".format(CONTEXT_MOUNT_ARTIFACTS_FORMAT)
CONTEXT_MOUNT_RUN_EVENTS_FORMAT = "{}/events".format(CONTEXT_MOUNT_ARTIFACTS_FORMAT)
CONTEXT_MOUNT_RUN_SYSTEM_RESOURCES_EVENTS_FORMAT = "{}/resources".format(
    CONTEXT_MOUNT_ARTIFACTS_FORMAT
)
CONTEXT_MOUNT_SHM = "/dev/shm"
CONTEXT_MOUNT_DOCKER = "/var/run/docker.sock"

CONTEXT_TMP_POLYAXON_PATH = get_temp_path(".polyaxon")
CONTEXT_USER_POLYAXON_PATH = get_project_path(".polyaxon")

CONTEXT_TMP_RUNS_ROOT_FORMAT = os.environ.get(
    ENV_KEYS_ARCHIVES_ROOT, "/tmp/plx/.runs/{}"
)
CONTEXT_ARCHIVES_ROOT = os.environ.get(ENV_KEYS_ARCHIVES_ROOT, "/tmp/plx/archives")
CONTEXT_ARTIFACTS_ROOT = os.environ.get(ENV_KEYS_ARTIFACTS_ROOT, "/tmp/plx/artifacts")
CONTEXT_OFFLINE_ROOT = os.environ.get(ENV_KEYS_OFFLINE_ROOT, "/tmp/plx/offline")
CONTEXT_OFFLINE_FORMAT = "{}/{{}}".format(CONTEXT_OFFLINE_ROOT)
CONTEXT_ARTIFACTS_FORMAT = "{}/{{}}".format(CONTEXT_ARTIFACTS_ROOT)

CONTEXTS_OUTPUTS_SUBPATH_FORMAT = "{}/outputs"
CONTEXTS_EVENTS_SUBPATH_FORMAT = "{}/events"
CONTEXTS_SYSTEM_RESOURCES_EVENTS_SUBPATH_FORMAT = "{}/resources"


def get_offline_base_path(entity_kind: str, path: Optional[str] = None) -> str:
    from polyaxon._schemas.lifecycle import V1ProjectFeature

    path = path or CONTEXT_OFFLINE_ROOT
    entity_kind = "run" if entity_kind == V1ProjectFeature.RUNTIME else entity_kind
    return "{}/{}s".format(path.rstrip("/"), get_enum_value(entity_kind))


def get_offline_path(
    entity_value: str, entity_kind: str, path: Optional[str] = None
) -> str:
    from polyaxon._schemas.lifecycle import V1ProjectFeature

    path = path or CONTEXT_OFFLINE_ROOT
    entity_kind = "run" if entity_kind == V1ProjectFeature.RUNTIME else entity_kind
    return "{}/{}s/{}".format(
        path.rstrip("/"), get_enum_value(entity_kind), entity_value
    )


def mount_sandbox(path: Optional[str] = None):
    global CONTEXT_OFFLINE_ROOT
    global CONTEXT_ARTIFACTS_ROOT
    global CONTEXT_OFFLINE_FORMAT
    global CONTEXT_ARTIFACTS_FORMAT

    path = path or CONTEXT_OFFLINE_ROOT
    CONTEXT_OFFLINE_ROOT = path
    CONTEXT_ARTIFACTS_ROOT = path
    CONTEXT_OFFLINE_FORMAT = "{}/{{}}".format(CONTEXT_OFFLINE_ROOT)
    CONTEXT_ARTIFACTS_FORMAT = "{}/{{}}".format(CONTEXT_ARTIFACTS_ROOT)
    return path



from typing import Any, Dict, List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_run_reference_catalog import V1RunReferenceCatalog
from polyaxon._sdk.schemas.v1_settings_catalog import V1SettingsCatalog


class V1RunSettings(BaseAllowSchemaModel):
    namespace: Optional[StrictStr]
    agent: Optional[V1SettingsCatalog]
    queue: Optional[V1SettingsCatalog]
    artifacts_store: Optional[V1SettingsCatalog]
    tensorboard: Optional[Dict[str, Any]]
    build: Optional[Dict[str, Any]]
    component: Optional[Dict[str, Any]]
    models: Optional[List[V1RunReferenceCatalog]]
    artifacts: Optional[List[V1RunReferenceCatalog]]



import os

from typing import Optional

from polyaxon._env_vars.keys import ENV_KEYS_SERVICE_MODE
from polyaxon._schemas.lifecycle import V1ProjectFeature
from polyaxon._services import PolyaxonServices


def get_store_path(store_path: str, subpath: str, entity: Optional[str] = None) -> str:
    full_path = store_path

    if os.environ.get(ENV_KEYS_SERVICE_MODE) == PolyaxonServices.VIEWER:
        dir_path = "runs" if entity == V1ProjectFeature.RUNTIME else f"{entity}s"
        full_path = os.path.join(full_path, dir_path)

    if subpath:
        full_path = os.path.join(full_path, subpath)
    return full_path



from clipped.config.parser import ConfigParser as _ConfigParser
from clipped.decorators.memoization import memoize


class ConfigParser(_ConfigParser):
    @staticmethod
    @memoize
    def type_mapping():
        from polyaxon import types

        return types.MAPPING

    @staticmethod
    @memoize
    def type_forwarding():
        from polyaxon import types

        return types.FORWARDING



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._schemas.lifecycle import V1StageCondition


class V1EntityStageBodyRequest(BaseAllowSchemaModel):
    owner: Optional[StrictStr]
    entity: Optional[StrictStr]
    kind: Optional[StrictStr]
    name: Optional[StrictStr]
    condition: Optional[V1StageCondition]



import os

from clipped.utils.http import clean_host

from polyaxon._env_vars.keys import ENV_KEYS_PLATFORM_HOST
from polyaxon.api import LOCALHOST


def get_api_host(default: str = LOCALHOST):
    return clean_host(os.environ.get(ENV_KEYS_PLATFORM_HOST, default))



from typing import Any, Dict, Optional

from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr


class V1RunEdgeLineage(BaseAllowSchemaModel):
    """
    V1RunEdgeLineage
    """

    uuid: Optional[UUIDStr]
    is_upstream: Optional[bool]
    values: Optional[Dict[str, Any]]



from typing import Optional

from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._schemas.version import V1Version


class V1Compatibility(BaseSchemaModel):
    cli: Optional[V1Version]
    platform: Optional[V1Version]
    agent: Optional[V1Version]
    ui: Optional[V1Version]



from polyaxon._flow.builds import V1Build
from polyaxon._flow.cache import V1Cache
from polyaxon._flow.component import V1Component
from polyaxon._flow.dags import DagOpSpec
from polyaxon._flow.early_stopping import (
    V1DiffStoppingPolicy,
    V1FailureEarlyStopping,
    V1MedianStoppingPolicy,
    V1MetricEarlyStopping,
    V1TruncationStoppingPolicy,
)
from polyaxon._flow.environment import V1Environment
from polyaxon._flow.events import V1EventKind, V1EventTrigger
from polyaxon._flow.hooks import V1Hook
from polyaxon._flow.init import V1Init
from polyaxon._flow.io import V1IO
from polyaxon._flow.joins import V1Join, V1JoinParam
from polyaxon._flow.matrix import (
    AcquisitionFunctions,
    GaussianProcessConfig,
    GaussianProcessesKernels,
    MatrixMixin,
    UtilityFunctionConfig,
    V1Bayes,
    V1GridSearch,
    V1HpChoice,
    V1HpDateRange,
    V1HpDateTimeRange,
    V1HpGeomSpace,
    V1HpLinSpace,
    V1HpLogNormal,
    V1HpLogSpace,
    V1HpLogUniform,
    V1HpNormal,
    V1HpPChoice,
    V1HpQLogNormal,
    V1HpQLogUniform,
    V1HpQNormal,
    V1HpQUniform,
    V1HpRange,
    V1HpUniform,
    V1Hyperband,
    V1Hyperopt,
    V1Iterative,
    V1Mapping,
    V1Matrix,
    V1MatrixKind,
    V1RandomSearch,
    V1Tuner,
    validate_pchoice,
)
from polyaxon._flow.mounts import V1ArtifactsMount
from polyaxon._flow.notifications import V1Notification
from polyaxon._flow.operations import V1CompiledOperation, V1Operation
from polyaxon._flow.optimization import (
    V1Optimization,
    V1OptimizationMetric,
    V1OptimizationResource,
    V1ResourceType,
)
from polyaxon._flow.params import ParamSpec, V1Param, ops_params
from polyaxon._flow.plugins import V1Plugins
from polyaxon._flow.references import RefMixin, V1DagRef, V1HubRef, V1PathRef, V1UrlRef
from polyaxon._flow.run import (
    MXJobMode,
    RunMixin,
    V1CleanerJob,
    V1CleanPodPolicy,
    V1CloningKind,
    V1Dag,
    V1DaskJob,
    V1DaskReplica,
    V1Job,
    V1KFReplica,
    V1MPIJob,
    V1MXJob,
    V1NotifierJob,
    V1PaddleElasticPolicy,
    V1PaddleJob,
    V1PipelineKind,
    V1PytorchElasticPolicy,
    V1PytorchJob,
    V1RayJob,
    V1RayReplica,
    V1RunEdgeKind,
    V1RunKind,
    V1RunPending,
    V1RunResources,
    V1Runtime,
    V1SchedulingPolicy,
    V1Service,
    V1TFJob,
    V1TunerJob,
    V1XGBoostJob,
    validate_run_patch,
)
from polyaxon._flow.schedules import (
    ScheduleMixin,
    V1CronSchedule,
    V1DateTimeSchedule,
    V1IntervalSchedule,
    V1ScheduleKind,
)
from polyaxon._flow.templates import V1Template
from polyaxon._flow.termination import V1Termination
from polyaxon._flow.trigger_policies import V1TriggerPolicy

# Forward references for operations and components
V1Dag.update_forward_refs(V1Operation=V1Operation)
V1Dag.update_forward_refs(V1Component=V1Component)



from typing import List, Optional, Union
from typing_extensions import Literal

from polyaxon._flow.component.base import BaseComponent
from polyaxon._flow.io import V1IO
from polyaxon._flow.references import RefMixin
from polyaxon._flow.run import RunMixin, V1Runtime
from polyaxon._flow.templates import TemplateMixinConfig, V1Template


class V1Component(
    BaseComponent,
    TemplateMixinConfig,
    RunMixin,
    RefMixin,
):
    """Component is a discrete, repeatable, and self-contained action that defines
    an environment and a runtime.

    A component is made of code that performs an action,
    such as container building, data preprocessing, data transformation, model training, and so on.

    You can use any language to write the logic of your component,
    Polyaxon uses containers to execute that logic.

    Components are definitions that can be shared if they reach a
    certain maturity and can be managed by the [Component Hub](/docs/management/component-hub/).
    This allows you to create a library of frequently-used components and reuse them
    either by submitting them directly or by referencing them from your operations.

    Components can be used as well to extract as much information and be used as templates
    with default queues, container resources requirements, node scheduling, ...

    Args:
        version: str
        kind: str, should be equal to `component`
        name: str, optional
        description: str, optional
        tags: List[str], optional
        presets: List[str], optional
        queue: str, optional
        namespace: str, optional
        cache: [V1Cache](/docs/automation/helpers/cache/), optional
        termination: [V1Termination](/docs/core/specification/termination/), optional
        plugins: [V1Plugins](/docs/core/specification/plugins/), optional
        build: [V1Build](/docs/automation/builds/), optional
        hooks: List[[V1Hook](/docs/automation/hooks/)], optional
        inputs: [V1IO](/docs/core/specification/io/), optional
        outputs: [V1IO](/docs/core/specification/io/), optional
        run: Union[[V1Job](/docs/experimentation/jobs/), [V1Service](/docs/experimentation/services/), [V1TFJob](/docs/experimentation/distributed/tf-jobs/), [V1PytorchJob](/docs/experimentation/distributed/pytorch-jobs/), [V1MPIJob](/docs/experimentation/distributed/mpi-jobs/), [V1PaddleJob](/docs/experimentation/distributed/paddle-jobs/), [V1RayJob](/docs/experimentation/distributed/ray-jobs/), [V1DaskJob](/docs/experimentation/distributed/dask-jobs/), [V1Dag](/docs/automation/flow-engine/specification/)]  # noqa
        template: [V1Template](/docs/core/specification/template/), optional

    ## YAML usage

    ```yaml
    >>> component:
    >>>   version: 1.1
    >>>   kind: component
    >>>   name:
    >>>   description:
    >>>   tags:
    >>>   presets:
    >>>   queue:
    >>>   namespace:
    >>>   cache:
    >>>   termination:
    >>>   plugins:
    >>>   actions:
    >>>   hooks:
    >>>   inputs:
    >>>   outputs:
    >>>   build:
    >>>   run:
    >>>   isApproved:
    >>>   template:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import (
    >>>     V1Build, V1Cache, V1Component, V1Hook, V1IO, V1Plugins, V1Termination
    >>> )
    >>> component = V1Component(
    >>>     name="test",
    >>>     description="test",
    >>>     tags=["test"],
    >>>     presets=["test"],
    >>>     queue="test",
    >>>     namespace="test",
    >>>     cache=V1Cache(...),
    >>>     termination=V1Termination(...),
    >>>     plugins=V1Plugins(...),
    >>>     hooks=[V1Hook(...)],
    >>>     inputs=[V1IO(...)],
    >>>     outputs=[V1IO(...)],
    >>>     build=V1Build(...),
    >>>     run=...
    >>> )
    ```

    ## Fields

    ### version

    The polyaxon specification version to use to validate the component.

    If you are using the component inline in an operation, this field is not required since it
    will be populated by the operation.

    ```yaml
    >>> component:
    >>>   version: 1.1
    ```

    ### kind

    The kind signals to the CLI, client, and other tools that this is a component.

    If you are using the component inline in an operation or a dag or
    if you are using the python client to create a component,
    this field is not required and is set by default.

    ```yaml
    >>> component:
    >>>   kind: component
    ```

    ### name

    The default component name.

    This name can be a `slug`, a `slug:tag`, `org/slug`, or `org/slug:slug`.

    This name will be passed as the default value to all operations using this component,
    unless the operations override the name or a `--name`
    is passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   name: test
    ```

    ### description

    The default component description.

    This description will be passed as the default value to all operations using this component,
    unless the operations override the description or a
    `--description` is passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   description: test
    ```

    ### tags

    The default component tags.

    These tags will be passed as the default value to all operations using this component,
    unless the operations override the tags or `--tags` are passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   tags: [test]
    ```

    ### presets

    The default component [presets](/docs/core/scheduling-presets/).

    These presets will be passed as the default value to all operations using this component,
    unless the operations override the presets or `--presets`
    is passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   presets: [test]
    ```

    ### queue

    The default component [queue](/docs/core/scheduling-strategies/queues/).

    This queue will be passed as the default value to all operations using this component,
    unless the operations override the queue or `--queue`
    is passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   queue: agent-name/queue-name
    ```

    If the agent name is not specified, Polyaxon will resolve the name of the queue
    based on the default agent.

    ```yaml
    >>> component:
    >>>   queue: queue-name
    ```

    ### namespace

    > **Note**: Please note that this field is only available in some commercial editions.

    The namespace to use, if not provided, it will default to the agent's namespace.

    ```yaml
    >>> component:
    >>>   namespace: polyaxon
    ```

    ### cache

    The default component [cache](/docs/automation/helpers/cache/).

    This cache definition will be passed as the default value to
    all operations using this component,
    unless the operations override the cache or `--nocache`
    is passed as an argument to the cli/client.

    ```yaml
    >>> component:
    >>>   cache:
    >>>     disable: false
    >>>     ttl: 100
    ```

    ### termination

    The default component [termination](/docs/core/specification/termination/).

    This termination definition will be passed as the default value to
    all operations using this component,
    unless the operations override the termination.

    ```yaml
    >>> component:
    >>>   termination:
    >>>     maxRetries: 2
    ```

    ### plugins

    The default component [plugins](/docs/core/specification/plugins/).

    This plugins definition will be passed as the default value to
    all operations using this component,
    unless the operations override the plugins.

    ```yaml
    >>> component:
    >>>   name: debug
    >>>   ...
    >>>   plugins:
    >>>     auth: false
    >>>     collectLogs: false
    >>>   ...
    ```

    Build using docker:

    ```yaml
    >>> component:
    >>>   name: build
    >>>   ...
    >>>   plugins:
    >>>     docker: true
    >>>   ...
    ```

    ### inputs

    The [inputs](/docs/core/specification/io/) definition for this component.

    If the component defines required inputs, anytime a user tries to run
    this component without passing the required params or passing params with wrong types,
    an exception will be raised.

    ```yaml
    >>> component:
    >>>   name: tensorboard
    >>>   ...
    >>>   inputs:
    >>>     - name: image
    >>>       type: str
    >>>       isOptional: true
    >>>       value: tensorflow:2.1
    >>>     - name: log_dir
    >>>       type: path
    >>>   ...
    ```

    ### outputs

    The [outputs](/docs/core/specification/io/) definition for this component.

    If the component defines required outputs, no exception will be raised at execution time,
    since Polyaxon considers the output values will be resolved in the future,
    for example during the run time when the user will be using the tracking
    client to log a metric or a value or an artifact.

    Sometimes the outputs can be resolved immediately at execution time,
    for example a container image name, because such information is required for the
    job to finish successfully, i.e. pushing the image with the correct name,
    in that case you can disable the `delayValidation` flag.

    ```yaml
    >>> component:
    >>>   name: tensorboard
    >>>   ...
    >>>   outputs:
    >>>     - name: image
    >>>       type: str
    >>>       delayValidation: false
    >>>   ...
    ```

    ### build

    > **Note**: Please check [V1Build](/docs/automation/builds/) for more details.

    This section defines if this component should build a container before starting the main logic.
    If the build section is provided, Polyaxon will set the main operation to a pending state
    until the build is done and then it will use the resulting docker image
    for starting the main container.

    ```yaml
    >>> component:
    >>>   ...
    >>>   build:
    >>>     hubRef: kaniko
    >>>   ...
    ```

    ### run

    This is the section that defines the runtime of the component:
     * [V1Job](/docs/experimentation/jobs/): for running batch jobs, model training experiments,
       data processing jobs, ...
     * [V1Service](/docs/experimentation/services/): for running tensorboards, notebooks,
       streamlit, custom services or an API.
     * [V1TFJob](/docs/experimentation/distributed/tf-jobs/): for running distributed
       Tensorflow training job.
     * [V1PytorchJob](/docs/experimentation/distributed/pytorch-jobs/): for running distributed
       Pytorch training job.
     * [V1MPIJob](/docs/experimentation/distributed/mpi-jobs/): for running distributed MPI job.
     * [V1PaddleJob](/docs/experimentation/distributed/paddle-jobs/): for running distributed
       PaddlePaddle training job.
     * [V1RayJob](/docs/experimentation/distributed/ray-jobs/): for running a ray job.
     * [V1DaskJob](/docs/experimentation/distributed/dask-jobs/): for running a Dask job.
     * [V1Dag](/docs/automation/flow-engine/specification/): for running a DAG/workflow.

    ### isApproved

    This is a flag to trigger human validation before queuing and scheduling this component.
    The default behavior is `True` even when the field is not set, i.e. no validation is required.
    To require a human validation prior to scheduling an operation,
    you can set this field to `False`.

    ```yaml
    >>> isApproved: false
    ```

    ### Cost

    A field to define the cost of running the operation. The value is a float and should map to a
    convention of a cost estimation in your team or
    it can map directly to the cost of using the environment where the operation is running.

    ```yaml
    >>> cost: 2.2
    ```
    """

    _IDENTIFIER = "component"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    inputs: Optional[List[V1IO]]
    outputs: Optional[List[V1IO]]
    run: Union[V1Runtime]
    template: Optional[V1Template]

    def get_run_kind(self):
        return self.run.kind if self.run else None

    def get_kind_value(self):
        return self.name

    def get_run_dict(self):
        config_dict = self.to_light_dict()
        config_dict.pop("tag", None)
        return config_dict

    def get_name(self):
        return self.name.split(":")[0] if self.name else None



from typing import Optional

from clipped.compact.pydantic import StrictStr

from polyaxon._schemas.base import BaseSchemaModel


class V1Version(BaseSchemaModel):
    min: Optional[StrictStr]
    latest: Optional[StrictStr]



from polyaxon import settings
from polyaxon.exceptions import PolyaxonClientException


def ensure_is_managed():
    if not settings.CLIENT_CONFIG.is_managed and not settings.CLIENT_CONFIG.in_cluster:
        raise PolyaxonClientException("This experiment/job is not managed by Polyaxon.")



from typing import Dict, List, Optional, Tuple

from clipped.utils.lists import to_list
from clipped.utils.sanitizers import sanitize_string_dict

from polyaxon._flow import V1Environment
from polyaxon._k8s import k8s_schemas
from polyaxon.exceptions import PolyaxonConverterError


def get_pod_spec(
    resource_name: str,
    namespace: str,
    main_container: k8s_schemas.V1Container,
    sidecar_containers: Optional[List[k8s_schemas.V1Container]],
    init_containers: Optional[List[k8s_schemas.V1Container]],
    environment: Optional[V1Environment],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    volumes: Optional[List[k8s_schemas.V1Volume]],
) -> Tuple[k8s_schemas.V1ObjectMeta, k8s_schemas.V1PodSpec]:
    if not main_container:
        raise PolyaxonConverterError("A main container is required")
    environment = environment or V1Environment()

    metadata = k8s_schemas.V1ObjectMeta(
        name=resource_name,
        namespace=namespace,
        labels=labels,
        annotations=annotations,
    )

    init_containers = to_list(init_containers, check_none=True)
    containers = [main_container] + to_list(sidecar_containers, check_none=True)
    image_pull_secrets = None
    if environment.image_pull_secrets:
        image_pull_secrets = [
            k8s_schemas.V1LocalObjectReference(name=i)
            for i in environment.image_pull_secrets
        ]

    pod_spec = k8s_schemas.V1PodSpec(
        init_containers=init_containers,
        containers=containers,
        volumes=volumes,
        restart_policy=environment.restart_policy,
        image_pull_secrets=image_pull_secrets,
        security_context=environment.security_context,
        service_account_name=environment.service_account_name,
        node_selector=sanitize_string_dict(environment.node_selector),
        tolerations=environment.tolerations,
        affinity=environment.affinity,
        dns_config=environment.dns_config,
        dns_policy=environment.dns_policy,
        host_aliases=environment.host_aliases,
        host_network=environment.host_network,
        host_pid=str(environment.host_pid)
        if environment.host_pid is not None
        else environment.host_pid,
        node_name=environment.node_name,
        priority=environment.priority,
        priority_class_name=environment.priority_class_name,
        scheduler_name=environment.scheduler_name,
    )
    return metadata, pod_spec


def get_pod_template_spec(
    metadata: k8s_schemas.V1ObjectMeta, pod_spec: k8s_schemas.V1PodSpec
) -> k8s_schemas.V1PodTemplateSpec:
    return k8s_schemas.V1PodTemplateSpec(metadata=metadata, spec=pod_spec)



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from clipped.compact.pydantic import (
    Field,
    PositiveInt,
    StrictInt,
    root_validator,
    validator,
)
from clipped.config.schema import skip_partial
from clipped.types.ref_or_obj import RefField

from polyaxon._flow.early_stopping import V1EarlyStopping
from polyaxon._flow.matrix.base import BaseSearchConfig
from polyaxon._flow.matrix.enums import (
    AcquisitionFunctions,
    GaussianProcessesKernels,
    V1MatrixKind,
)
from polyaxon._flow.matrix.params import V1HpParam
from polyaxon._flow.matrix.tuner import V1Tuner
from polyaxon._flow.optimization import V1OptimizationMetric
from polyaxon._schemas.base import BaseSchemaModel


class GaussianProcessConfig(BaseSchemaModel):
    _IDENTIFIER = "gaussian_process"

    kernel: Optional[GaussianProcessesKernels] = Field(
        default=GaussianProcessesKernels.MATERN
    )
    length_scale: Optional[float] = Field(default=1.0, alias="lengthScale")
    nu: Optional[float] = Field(default=1.5)
    num_restarts_optimizer: Optional[int] = Field(
        default=0, alias="numRestartsOptimizer"
    )


def validate_utility_function(acquisition_function, kappa, eps):
    condition = AcquisitionFunctions.is_ucb(acquisition_function) and kappa is None
    if condition:
        raise ValueError("the acquisition function `ucb` requires a parameter `kappa`")

    condition = (
        AcquisitionFunctions.is_ei(acquisition_function)
        or AcquisitionFunctions.is_poi(acquisition_function)
    ) and eps is None
    if condition:
        raise ValueError(
            "the acquisition function `{}` requires a parameter `eps`".format(
                acquisition_function
            )
        )


class UtilityFunctionConfig(BaseSchemaModel):  # TODO: Rename to V1UtilityFunction
    _IDENTIFIER = "utility_function"

    acquisition_function: Optional[AcquisitionFunctions] = Field(
        default=AcquisitionFunctions.UCB, alias="acquisitionFunction"
    )
    gaussian_process: Optional[GaussianProcessConfig] = Field(alias="gaussianProcess")
    kappa: Optional[float]
    eps: Optional[float]
    num_warmup: Optional[int] = Field(alias="numWarmup")
    num_iterations: Optional[int] = Field(alias="numIterations")

    @root_validator
    @skip_partial
    def validate_utility_function(cls, values):
        validate_utility_function(
            acquisition_function=values.get("acquisition_function"),
            kappa=values.get("kappa"),
            eps=values.get("eps"),
        )
        return values


def validate_matrix(matrix):
    if not matrix:
        return None

    for key, value in matrix.items():
        if value.is_distribution and not value.is_uniform:
            raise ValueError(
                "`{}` defines a non uniform distribution, "
                "and it cannot be used with bayesian optimization.".format(key)
            )

    return matrix


class V1Bayes(BaseSearchConfig):
    """Bayesian optimization is an extremely powerful technique.
    The main idea behind it is to compute a posterior distribution
    over the objective function based on the data, and then select good points
    to try with respect to this distribution.

    The way Polyaxon performs bayesian optimization is by measuring
    the expected increase in the maximum objective value seen over all
    experiments in the group, given the next point we pick.

    Args:
        kind: string, should be equal to `bayes`
        utility_function: UtilityFunctionConfig
        num_initial_runs: int
        max_iterations: int
        metric: V1OptimizationMetric
        params: List[Dict[str, [params](/docs/automation/optimization-engine/params/#discrete-values)]]  # noqa
        seed: int, optional
        concurrency: int, optional
        tuner: [V1Tuner](/docs/automation/optimization-engine/tuner/), optional
        early_stopping: List[[EarlyStopping](/docs/automation/helpers/early-stopping)], optional


    ## YAML usage

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   utilityFunction:
    >>>   numInitialRuns:
    >>>   maxIterations:
    >>>   metric:
    >>>   params:
    >>>   seed:
    >>>   concurrency:
    >>>   tuner:
    >>>   earlyStopping:
    ```

    ## Python usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import (
    >>>     V1Bayes, V1HpLogSpace, V1HpChoice, V1FailureEarlyStopping, V1MetricEarlyStopping,
    >>>     V1OptimizationMetric, V1Optimization, V1OptimizationResource, UtilityFunctionConfig
    >>> )
    >>> matrix = V1Bayes(
    >>>   concurrency=20,
    >>>   utility_function=UtilityFunctionConfig(...),
    >>>   num_initial_runs=40,
    >>>   max_iterations=20,
    >>>   params={"param1": V1HpLogSpace(...), "param2": V1HpChoice(...), ... },
    >>>   metric=V1OptimizationMetric(name="loss", optimization=V1Optimization.MINIMIZE),
    >>>   early_stopping=[V1FailureEarlyStopping(...), V1MetricEarlyStopping(...)]
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this matrix is bayes.

    If you are using the python client to create the mapping,
    this field is not required and is set by default.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    ```

    ### params

    A dictionary of `key -> value generator`
    to generate the parameters.

    To learn about all possible
    [params generators](/docs/automation/optimization-engine/params/).

    > The parameters generated will be validated against
    > the component's inputs/outputs definition to check that the values
    > can be passed and have valid types.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   params:
    >>>     param1:
    >>>        kind: ...
    >>>        value: ...
    >>>     param2:
    >>>        kind: ...
    >>>        value: ...
    ```

    ### utilityFunction

    the utility function defines what acquisition function and bayesian process to use.

    ### Acquisition functions

    A couple of acquisition functions can be used: `ucb`, `ei` or `poi`.

     * `ucb`: Upper Confidence Bound,
     * `ei`: Expected Improvement
     * `poi`: Probability of Improvement

    When using `ucb` as acquisition function, a tunable parameter `kappa`
    is also required, to balance exploitation against exploration, increasing kappa
    will make the optimized hyperparameters pursuing exploration.

    When using `ei` or `poi` as acquisition function, a tunable parameter `eps` is also required,
    to balance exploitation against exploration, increasing epsilon will
    make the optimized hyperparameters more spread out across the whole range.

    ### Gaussian process

    Polyaxon allows to tune the gaussian process.

     * `kernel`: `matern` or `rbf`.
     * `lengthScale`: float
     * `nu`: float
     * `numRestartsOptimizer`: int

     ```yaml
     >>> matrix:
     >>>   kind: bayes
     >>>   utility_function:
     >>>     acquisitionFunction: ucb
     >>>     kappa: 1.2
     >>>     gaussianProcess:
     >>>       kernel: matern
     >>>       lengthScale: 1.0
     >>>       nu: 1.9
     >>>       numRestartsOptimizer: 0
     ```

    ### numInitialRuns

    the initial iteration of random experiments is required to create a seed of observations.

    This initial random results are used by the algorithm to update
    the regression model for generating the next suggestions.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   numInitialRuns: 40
    ```

    ### maxIterations

    After creating the first set of random observations,
    the algorithm will use these results to update
    the regression model and suggest a new experiment to run.

    Every time an experiment is done,
    the results are used as an observation and are appended
    to the historical values so that the algorithm can use all
    the observations again to suggest more experiments to run.

    The algorithm will keep suggesting more experiments and adding
    their results as an observation, every time we make a new observation,
    i.e. an experiment finishes and reports the results to the platform,
    the results are appended to the historical values, and then used to make a better suggestion.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   maxIterations: 15
    ```

    This configuration will make 15 suggestions based on the historical values,
    every time an observation is made is appended to the historical values
    to make better subsequent suggestions.

    ### metric

    The metric to optimize during the iterations,
    this is the metric that you want to maximize or minimize.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   metric:
    >>>     name: loss
    >>>     optimization: minimize
    ```

    ### seed

    Since this algorithm uses random generators,
    if you want to control the seed for the random generator, you can pass a seed.

     ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   seed: 523
    ```

    ### concurrency

    An optional value to set the number of concurrent operations.

    <blockquote class="light">
    This value only makes sense if less or equal to the total number of possible runs.
    </blockquote>

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   concurrency: 20
    ```

    For more details about concurrency management,
    please check the [concurrency section](/docs/automation/helpers/concurrency/).

    ### earlyStopping

    A list of early stopping conditions to check for terminating
    all operations managed by the pipeline.
    If one of the early stopping conditions is met,
    a signal will be sent to terminate all running and pending operations.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   earlyStopping: ...
    ```

    For more details please check the
    [early stopping section](/docs/automation/helpers/early-stopping/).

    ### tuner

    The tuner reference (w/o component hub reference) to use.
    The component contains the logic for creating new suggestions based on bayesian optimization,
    users can override this section to provide a different tuner component.

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   tuner:
    >>>     hubRef: 'acme/my-bo-tuner:version'
    ```

    ## Example

    This is an example of using bayesian search for hyperparameter tuning:

    ```yaml
    >>> matrix:
    >>>   kind: bayes
    >>>   concurrency: 5
    >>>   maxIterations: 15
    >>>   numInitialTrials: 30
    >>>   metric:
    >>>     name: loss
    >>>     optimization: minimize
    >>>   utilityFunction:
    >>>     acquisitionFunction: ucb
    >>>     kappa: 1.2
    >>>     gaussianProcess:
    >>>       kernel: matern
    >>>       lengthScale: 1.0
    >>>       nu: 1.9
    >>>       numRestartsOptimizer: 0
    >>>   params:
    >>>     lr:
    >>>       kind: uniform
    >>>       value: [0, 0.9]
    >>>     dropout:
    >>>       kind: choice
    >>>       value: [0.25, 0.3]
    >>>     activation:
    >>>       kind: pchoice
    >>>       value: [[relu, 0.1], [sigmoid, 0.8]]
    >>> component:
    >>>   inputs:
    >>>     - name: batch_size
    >>>       type: int
    >>>       isOptional: true
    >>>       value: 128
    >>>     - name: lr
    >>>       type: float
    >>>     - name: dropout
    >>>       type: float
    >>>   container:
    >>>     image: image:latest
    >>>     command: [python3, train.py]
    >>>     args: [
    >>>         "--batch-size={{ batch_size }}",
    >>>         "--lr={{ lr }}",
    >>>         "--dropout={{ dropout }}",
    >>>         "--activation={{ activation }}"
    ```
    """

    _IDENTIFIER = V1MatrixKind.BAYES

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    utility_function: Optional[UtilityFunctionConfig] = Field(alias="utilityFunction")
    num_initial_runs: Union[PositiveInt, RefField] = Field(alias="numInitialRuns")
    max_iterations: Union[PositiveInt, RefField] = Field(alias="maxIterations")
    metric: V1OptimizationMetric
    params: Union[Dict[str, V1HpParam], RefField]
    seed: Optional[Union[StrictInt, RefField]]
    concurrency: Optional[Union[PositiveInt, RefField]]
    tuner: Optional[Union[V1Tuner, RefField]]
    early_stopping: Optional[Union[List[V1EarlyStopping], RefField]] = Field(
        alias="earlyStopping"
    )

    def create_iteration(self, iteration: Optional[int] = None) -> int:
        if iteration is None:
            return 0
        return iteration + 1

    def should_reschedule(self, iteration):
        """Return a boolean to indicate if we need to reschedule another iteration."""
        return iteration < self.max_iterations

    @validator("params", always=True)
    @skip_partial
    def validate_matrix(cls, params):
        return validate_matrix(params)



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.types.email import EmailStr

from polyaxon._schemas.base import BaseSchemaModel


class UserConfig(BaseSchemaModel):
    _IDENTIFIER = "user"

    username: StrictStr
    email: Optional[EmailStr]
    name: Optional[StrictStr]
    theme: Optional[int]
